package org.dbgen;

import java.util.*;
import java.io.*;

/**
 * Given a table defintion (Table object) this class can be used to generate Java
 * source codes that binds directly to the data source using JDBC.  This is so-called
 * a Object-Relational binding.
 *
 * As of DbGen Professional Edition 1.0b2, this engine is now customizable should
 * the generated source codes need to be modified in certain programming environments.
 * Please see separate documentation on how to customize the DbGen engine.
 *
 */
public class DbGen
{
  protected transient Output output = new Output();
  java.io.PrintWriter fieldWriter = null;
  protected transient String NL = Output.NEWLINE;
  public static String PROGRAM = "DbGen";
  public static String VERSION = "1.2";
  public static String NAME = PROGRAM + " Version " + VERSION;
  public static String GENERATE_CREATE_SCRIPT = "GenerateCreateScript";
  public static String GENERATE_DATA_CLASS = "GenerateDataClass";
  public static String GENERATE_TEST_CLASS = "GenerateTestClass";
  public static String GENERATE_TABLE_MODEL = "GenerateTableModel";
  public static String GENERATE_TABLE_MODEL_JAVAX = "GenerateTableModelJavax";
  public static String GENERATE_TABLE_MODEL_MAIN = "GenerateTableModelMain";
  public static String CREATE_PACKAGE_SUBDIRECTORIES = "CreatePackageSubDirectories";
  public static String DATABASE_TYPE = "DatabaseType";
  public final static int DATABASE_ORACLE = 1;
  public final static int DATABASE_SYBASE = 2;
  public final static int DATABASE_MSQL = 3;
  public final static int DATABASE_ACCESS = 4;
  public final static int DATABASE_SQLSERVER = 5;
  protected boolean msqlTrick = false;

  /**
   * Make sure the output directory (including the package sub-directory path) exist.
   * @return Real output directory.
   * @param table The Table object.
   * @param outputDirectory The output directory.
   */
  public String ensurePackagePath(Project project, Table table, String outputDirectory) throws IOException
  {
    /* Retrieve default package name and table override package name. */
    String packageName = getPackageName(project, table);

    /* Just return the current output directory if there is no package name specified. */
    if (packageName == null || org.dbgen.support.StringHelper.isBlank(packageName))
      return outputDirectory;

    /* Since there is a package name, decode the name and create necessary directories. */
    StringTokenizer st = new StringTokenizer(packageName, ".");
    String currentDirectory = outputDirectory;

    while (st.hasMoreTokens())
    {
      currentDirectory += (File.separator + st.nextToken());

      //org.dbgen.Debug.println("Checking directory " + currentDirectory + "...");
      File file = new File(currentDirectory);

      if (!file.exists())
      {
        boolean good = file.mkdir();

        if (!good)
        {
          throw new IOException("Directory " + currentDirectory + " cannot be created.");
        }

        org.dbgen.Debug.println("Directory " + currentDirectory + " created.");
      }
      else
        if (!file.isDirectory())
      {
        throw new IOException("The path " + currentDirectory + " is not a directory.");
      }
    }

    return currentDirectory;
  }

  /**
   * This method was created by a SmartGuide.
   */
  protected void flushWriter()
  {
    fieldWriter.close();

    return;
  }

  /**
   * Generate Java source codes (Data and Test) for the specified table.
   * @param project The Project object.
   * @param table The Table object.
   * @param options Generation options.
   */
  public void generate(Project project, Table table, Hashtable options) throws IOException
  {
    String outputDirectory = project.getOutputDirectory();
    String path = outputDirectory;

    /* By default we will create sub directories */
    Boolean createSubDir = (Boolean) options.get(CREATE_PACKAGE_SUBDIRECTORIES);

    if (createSubDir == null || createSubDir.booleanValue())
    {
      path = ensurePackagePath(project, table, outputDirectory);
    }

    /* Database specific tricks */
    int dbtype = ((Integer) options.get(DATABASE_TYPE)).intValue();
    msqlTrick = (dbtype == DATABASE_MSQL);

    /* Optional generation */
    Boolean go;
    go = (Boolean) options.get(GENERATE_DATA_CLASS);

    if (go != null && go.booleanValue())
    {
      String dataFilename = table.getClassName() + ".java";
      setWriter(new PrintWriter(new FileWriter(path + File.separator + dataFilename)));
      generateDataClass(project, table);
      flushWriter();
    }

    go = (Boolean) options.get(GENERATE_TEST_CLASS);

    if (go != null && go.booleanValue())
    {
      String testFilename = table.getClassName() + "Test.java";
      setWriter(new PrintWriter(new FileWriter(path + File.separator + testFilename)));
      generateTestClass(project, table);
      flushWriter();
    }

    go = (Boolean) options.get(GENERATE_CREATE_SCRIPT);

    if (go != null && go.booleanValue())
    {
      String sqlFilename = table.getClassName() + ".sql";
      setWriter(new PrintWriter(new FileWriter(path + File.separator + sqlFilename)));
      generateCreateScript(table, dbtype);
      flushWriter();
    }

    go = (Boolean) options.get(GENERATE_TABLE_MODEL);

    if (go != null && go.booleanValue())
    {
      String tmFilename = table.getClassName() + "DbGenTableModel.java";
      setWriter(new PrintWriter(new FileWriter(path + File.separator + tmFilename)));
      generateTableModel(
        project,
        table,
        ((Boolean) options.get(GENERATE_TABLE_MODEL_JAVAX)).booleanValue(),
        ((Boolean) options.get(GENERATE_TABLE_MODEL_MAIN)).booleanValue()
        );

      flushWriter();
    }

    return;
  }

  /**
   * This method was created by a SmartGuide.
   * @param table org.dbgen.Table
   */
  public void generateCreateScript(Table table, int dbtype)
  {
    String goStr = (dbtype == DATABASE_MSQL ? "\\g" : "go");
    output.println("DROP TABLE " + table.getTableName());
    output.println(goStr);
    output.println();
    output.println("CREATE TABLE " + table.getTableName());
    output.println("(");
    Vector fields = table.getFields();
    int len = fields.size();

    for (int i = 0; i < len; i++)
    {
      Field f = (Field) fields.elementAt(i);

      if (i > 0)
      {
        output.println(",");
      }

      output.iprint(f.getName() + " " + getSQLType(f, dbtype));

      if (!f.isBooleanType())
      {
        output.print(f.getIsKey() ? " NOT NULL" : " NULL");
      }
    }

    if (dbtype != DATABASE_MSQL)
    {    // mSQL is special... not sql compatible...
      output.println(",");
      output.iprint("constraint " + table.getTableName() + "_keys0 primary key (");

      for (int i = 0; i < len; i++)
      {
        Field f = (Field) fields.elementAt(i);

        if (f.getIsKey())
        {
          if (i > 0)
          {
            output.print(", ");
          }

          output.print(f.getName());
        }
      }

      output.println(")");
    }

    output.println(")");
    output.println(goStr);
    output.println();

    return;
  }

  /**
   * Generate the Data object for this table.
   * @param table org.dbgen.Table
   */
  public void generateDataClass(Project project, Table table)
  {
    makeHeader(table);
    makeImport(project, table);
    output.println();
    output.println("public class " + table.getClassName() + " implements java.io.Serializable, " + getCustomDataClassName());
    output.println("{");
    output.indent();
    makeFieldDefinitions(table);
    makeConstructors(table);
    makeGetSetMethods(table);
    makePropertyChangeMethods();
    makeClearMethod(table);
    makeToStringMethod(table);
    makeGetTableNameMethod(table);
    makeGetKeyConditionsMethod(table);
    makeGetKeyConditionsForRelationMethod(table);
    makeGetInsertValuesMethod(table);
    makeGetUpdateValuesMethod(table);
    makeMakeObjectMethod(table);

    //	makeSetKeysFromParentMethod(project, table);

    makeCustomMethods(table);
    output.deindent();
    output.println("}");

    return;
  }

  /**
   * Generate the Data object for this table.
   * @param table org.dbgen.Table
   */
  public void generateTableModel(Project project, Table table, boolean isJavax, boolean isMain )
  {
    makeTMHeader(table );
    makeTMImport(project, table, isJavax);
    output.println();
    output.println("public class " + table.getClassName() + "DbGenTableModel extends org.dbgen.support.AbstractDbGenTableModel");
    output.println("{");
    output.indent();
    makeTMFieldDefinitions(table);
    makeTMConstructors(table);
    makeTMGetSetMethods(table);

    if ( isMain )
      makeTMMain(table);

    output.deindent();
    output.println("}");

    return;
  }

  /**
   * Generate the Test program for this table.
   * @param table org.dbgen.Table
   */
  public void generateTestClass(Project project, Table table)
  {
    String className = table.getClassName();
    Vector fields = table.getFields();
    Vector foreignTables = table.getForeignTablesWithRelation();

    // ----- Header -----
    output.comment(true, "Class " + className + "Test.java" + NL + NL + "This program is a front-end tool to modify the '" + table.getTableName() + "' table. " + "It currently supports LIST, INSERT, and DELETE functions. The general rule " + "to run the program looks like this: " + NL + NL + " 1. java progname list <condition> <orderby>" + NL + " 2. java progname insert <field1> <field2> ... <fieldN> " + NL + " 3. java progname delete <keyfield1> <keyfield2> ... <keyfieldN>" + NL + NL + "NOTE: Make sure you modify the JDBC Driver and URL variables before you try this program." + NL + NL + "Generated by: " + NAME + NL + "Generated on: " + Util.getDateString());

    /* Retrieve default package name and table override package name. */
    String packageName = getPackageName(project, table);

    //System.out.println("package name for test is " + packageName);
    if (packageName != null && !org.dbgen.support.StringHelper.isBlank(packageName))
    {
      output.println("package " + packageName + ";");
    }

    output.println();
    output.println("import java.sql.*;");
    output.println("import org.dbgen.support.*;");

    // ----- Start Class -----
    output.println();
    output.println("public class " + className + "Test {");
    output.indent();

    // ----- Main method -----
    output.iprintln("public static void main(String args[]) {");
    output.indent();
    output.iprintln("try {");
    output.iprintln(" // -- Select the right driver and url -- ");
    output.iprintln(" //String drv = \"sun.jdbc.odbc.JdbcOdbcDriver\";");
    output.iprintln(" //String url = \"jdbc:odbc:dsn\";");
    output.iprintln(" //String drv = \"com.imaginary.sql.msql.MsqlDriver\";");
    output.iprintln(" //String url = \"jdbc:msql://host:port/database\";");
    output.iprintln(" String drv = \"" + DbGenPro.getValue(DbGenPro.JDBC_DRIVER) + "\";");
    output.iprintln(" String url = \"" + DbGenPro.getValue(DbGenPro.JDBC_URL) + "\";");
    output.iprintln(" org.dbgen.support.DbAccess dba = new org.dbgen.support.DbAccess(drv, url, \"" + DbGenPro.getValue(DbGenPro.JDBC_USERNAME) + "\", \"" + DbGenPro.getValue(DbGenPro.JDBC_PASSWORD) + "\");");
    output.iprintln(" org.dbgen.support.Driver driver = new org.dbgen.support.Driver(dba);");
    output.iprintln(" dba.setDebug(true); // Set to TRUE for verbose SQL inspection");
    output.iprintln(" if (args[0].equalsIgnoreCase(\"insert\")) {");
    output.iprintln(" insert(driver, args);");
    output.iprintln(" } else if (args[0].equalsIgnoreCase(\"delete\")) {");
    output.iprintln(" delete(driver, args);");
    output.iprintln(" } else if (args[0].equalsIgnoreCase(\"list\")) {");
    output.iprintln(" list(driver, args);");
    output.iprintln(" } else if (args[0].equalsIgnoreCase(\"listd\")) {");

    if (foreignTables.size() > 0)
    {
      output.iprintln(" listDetails(driver, args);");
    }
    else
    {
      output.iprintln(" System.out.println(\"No foreign key defined.\");");
    }

    output.iprintln(" }");
    output.iprintln("} catch (Exception e) {");
    output.iprintln(" e.printStackTrace();");
    output.iprintln("}");
    output.deindent();
    output.iprintln("}");


    // ----- insert method -----
    output.iprintln("");
    output.iprintln("private static void insert(org.dbgen.support.Driver driver, String args[])");
    output.iprintln(" throws SQLException");
    output.iprintln("{");
    output.iprint(" " + className + " obj = new " + className + "(");

    for (int i = 0; i < fields.size(); i++)
    {
      Field f = (Field) fields.elementAt(i);

      if (i > 0)
      {
        output.print(", ");
      }

      int j = i + 1;

      if (f.isStringType())
      {
        output.print("args[" + j + "]");
      }
      else
      {
        output.print(getStringParseMethod(f) + "(args[" + j + "])");
      }
    }

    output.println(");");
    output.iprintln(" driver.insert(obj);");
    output.iprintln("}");

    // ----- delete method -----
    output.iprintln("");
    output.iprintln("private static void delete(org.dbgen.support.Driver driver, String args[])");
    output.iprintln(" throws SQLException");
    output.iprintln("{");
    output.iprint(" " + className + " obj = new " + className + "(");

    for (int i = 0, j = 0; i < fields.size(); i++)
    {
      Field f = (Field) fields.elementAt(i);

      if (f.getIsKey())
      {
        if (j++ > 0)
        {
          output.print(", ");
        }

        if (f.isStringType())
        {
          output.print("args[" + j + "]");
        }
        else
        {
          output.print(getStringParseMethod(f) + "(args[" + j + "])");
        }
      }
    }

    output.println(");");
    output.iprintln(" driver.delete(obj);");
    output.iprintln("}");

    // ----- list method -----
    output.iprintln("");
    output.iprintln("private static void list(org.dbgen.support.Driver driver, String args[])");
    output.iprintln(" throws SQLException");
    output.iprintln("{");
    output.iprintln(" " + className + " holder = new " + className + "();");
    output.iprintln(" ResultSet rs = driver.query(holder, args[1], args[2]);");
    output.iprintln(" " + className + " obj;");
    output.iprintln(" int i = 0;");
    output.iprintln(" while ((obj = (" + className + ") driver.getNextRecord(holder, rs)) != null) {");
    output.iprintln(" System.out.println((++i) + \".\" + obj.toString());");
    output.iprintln(" }");
    output.iprintln("}");


    // ----- list detail method -----
    output.iprintln("");
    output.iprintln("private static void list(org.dbgen.support.Driver driver, String args[])");
    output.iprintln(" throws SQLException");
    output.iprintln("{");
    output.iprintln(" " + className + " holder = new " + className + "();");
    output.iprintln(" ResultSet rs = driver.query(holder, args[1], args[2]);");
    output.iprintln(" " + className + " obj;");
    output.iprintln(" int i = 0;");
    output.iprintln(" while ((obj = (" + className + ") driver.getNextRecord(holder, rs)) != null) {");
    output.iprintln(" System.out.println((++i) + \".\" + obj.toString());");
    output.iprintln(" }");
    output.iprintln("}");

    if (foreignTables.size() > 0)
    {
      output.iprintln("");
      output.iprintln("private static void listDetails(org.dbgen.support.Driver driver, String args[])");
      output.iprintln(" throws SQLException");
      output.iprintln("{");
      output.iprintln(" " + className + " holder = new " + className + "();");
      output.iprintln(" ResultSet rs = driver.query(holder, args[1], args[2]);");
      output.iprintln(" " + className + " obj;");
      output.iprintln(" int i = 0;");
      output.iprintln(" while ((obj = (" + className + ") driver.getNextRecord(holder, rs)) != null) {");
      output.iprintln(" System.out.println();");
      output.iprintln(" System.out.println( \"===============================================================\" );\n");
      output.iprintln(" System.out.println((++i) + \".\" + obj.toString());");
      Table ft = null;

      for (int i = 0; i < foreignTables.size(); i++)
      {
        ft = (Table) foreignTables.elementAt(i);
        output.iprintln(" // For foreign table class " + ft.getClassName());
        output.iprintln(" System.out.println();");
        output.iprintln(" System.out.println( \"" + ft.getClassName() + ":\" );\n");
        output.iprintln(" try {");
        output.iprintln(" java.util.Vector details = driver.retrieveDetails( obj, new " + ft.getClassName() + "() );");
        output.iprintln(" if ( details != null ) {");
        output.iprintln(" for ( int j=0; j<details.size(); j++ )");
        output.iprintln(" System.out.println( \"\t\" + (j+1) + \".\" + details.elementAt(j).toString());");
        output.iprintln(" }");
        output.iprintln(" } catch ( Exception ex ) {");
        output.iprintln(" ex.printStackTrace();");
        output.iprintln(" }");
      }

      output.iprintln(" }");
      output.iprintln("}");
    }

    // ----- parseFloat method -----
    /*
     output.iprintln("");
     output.iprintln("private static float parseFloat(String s)");
     output.iprintln("{");
     output.iprintln("  float f = (float) 0.0;");
     output.iprintln("  try {");
     output.iprintln("    f = Float.valueOf(s).floatValue();");
     output.iprintln("  } catch (NumberFormatException e) {");
     output.iprintln("    e.printStackTrace();");
     output.iprintln("  }");
     output.iprintln("  return f;");
     output.iprintln("}");
     */

    // ----- parseDouble method -----
    /*
     output.iprintln("");
     output.iprintln("private static double parseDouble(String s)");
     output.iprintln("{");
     output.iprintln("  double f = 0.0;");
     output.iprintln("  try {");
     output.iprintln("    f = Double.valueOf(s).doubleValue();");
     output.iprintln("  } catch (NumberFormatException e) {");
     output.iprintln("    e.printStackTrace();");
     output.iprintln("  }");
     output.iprintln("  return f;");
     output.iprintln("}");
     */

    // ----- End Class -----
    output.iprintln("");
    output.deindent();
    output.println("}");
  }

  /**
   * Return the Data class name (customized Data object should override this method.)
   * @return java.lang.String
   */
  protected String getCustomDataClassName()
  {
    return "org.dbgen.support.Data";
  }

  /**
   * This method was created by a SmartGuide.
   * @return java.lang.String
   * @param field org.dbgen.Field
   */
  protected String getDatabaseAccessType(Field field)
  {
    switch (field.getType())
    {
    case java.sql.Types.BIT :

      return "BIT";

    case java.sql.Types.TINYINT :

      return "BYTE";

    case java.sql.Types.SMALLINT :

      return "SHORT";

    case java.sql.Types.INTEGER :

      return "LONG";

    case java.sql.Types.REAL :

      return "SINGLE";

    case java.sql.Types.FLOAT :

      return "DOUBLE";

    case java.sql.Types.DOUBLE :

      return "DOUBLE";

    case java.sql.Types.LONGVARCHAR :

      return "MEMO";

    case java.sql.Types.BINARY :

      return "UNSUPPORTED TYPE BINARY";

    case java.sql.Types.VARBINARY :

      return "UNSUPPORTED TYPE VARBINARY";

    case java.sql.Types.LONGVARBINARY :

      return "UNSUPPORTED TYPE LONGVARBINARY";

    case java.sql.Types.DATE :

    case java.sql.Types.TIME :

    case java.sql.Types.TIMESTAMP :

      return "DATETIME";

    case java.sql.Types.CHAR :

    case java.sql.Types.VARCHAR :

      return "TEXT(" + field.getLength() + ")";

    case java.sql.Types.DECIMAL :

      return "CURRENCY";

    case java.sql.Types.NUMERIC :

      return "CURRENCY";

    case java.sql.Types.OTHER :

      return null;
    }

    return null;
  }

  /**
   * This method was created by a SmartGuide.
   * @return java.lang.String
   * @param field org.dbgen.Field
   */
  protected String getDatabaseMsqlType(Field field)
  {
    switch (field.getType())
    {
    case java.sql.Types.BIT :

    case java.sql.Types.TINYINT :

    case java.sql.Types.SMALLINT :

    case java.sql.Types.INTEGER :

      return "int";

    case java.sql.Types.REAL :

    case java.sql.Types.FLOAT :

    case java.sql.Types.DOUBLE :

      return "real";

    case java.sql.Types.LONGVARCHAR :

      return "text(2048)";

    case java.sql.Types.BINARY :

    case java.sql.Types.VARBINARY :

      return "char(" + field.getLength() + ")";

    case java.sql.Types.LONGVARBINARY :

      return "text(2048)";

    case java.sql.Types.DATE :

    case java.sql.Types.TIME :

    case java.sql.Types.TIMESTAMP :

      // According to mSQL-JDBC_1.0b3 PROGRAMMING documentation
      return "char(21)";

    case java.sql.Types.CHAR :

    case java.sql.Types.VARCHAR :

      return "char(" + field.getLength() + ")";

    case java.sql.Types.DECIMAL :

    case java.sql.Types.NUMERIC :

      return "real";

    case java.sql.Types.OTHER :

      return null;
    }

    return null;
  }

  /**
   * This method was created by a SmartGuide.
   * @return java.lang.String
   * @param field org.dbgen.Field
   */
  protected String getDatabaseOracleType(Field field)
  {
    switch (field.getType())
    {
    case java.sql.Types.BIT :

      return "NUMBER(1)";

    case java.sql.Types.TINYINT :

      return "NUMBER(4)";

    case java.sql.Types.SMALLINT :

      return "NUMBER(6)";

    case java.sql.Types.INTEGER :

      return "NUMBER(12)";

    case java.sql.Types.REAL :

    case java.sql.Types.FLOAT :

    case java.sql.Types.DOUBLE :

      return "NUMBER";

    case java.sql.Types.LONGVARCHAR :

      return "LONG";

    case java.sql.Types.BINARY :

    case java.sql.Types.VARBINARY :

      return "RAW(" + field.getLength() + ")";

    case java.sql.Types.LONGVARBINARY :

      return "RAW";

    case java.sql.Types.DATE :

    case java.sql.Types.TIME :

    case java.sql.Types.TIMESTAMP :

      return "DATE";

    case java.sql.Types.CHAR :

      return "CHAR(" + field.getLength() + ")";

    case java.sql.Types.VARCHAR :

      return "VARCHAR2(" + field.getLength() + ")";

    case java.sql.Types.DECIMAL :

    case java.sql.Types.NUMERIC :

      return "NUMBER(" + field.getLength() + ", " + field.getScale() + ")";

    case java.sql.Types.OTHER :

      return null;
    }

    return null;
  }

  /**
   * This method was created by a SmartGuide.
   * @return java.lang.String
   * @param field org.dbgen.Field
   */
  protected String getDatabaseSqlServerType(Field field)
  {
    switch (field.getType())
    {
    case java.sql.Types.BIT :

      return "BIT";

    case java.sql.Types.TINYINT :

      return "TINYINT";

    case java.sql.Types.SMALLINT :

      return "SMALLINT";

    case java.sql.Types.INTEGER :

      return "INT";

    case java.sql.Types.REAL :

      return "REAL";

    case java.sql.Types.FLOAT :

      return "FLOAT";

    case java.sql.Types.DOUBLE :

      return "DOUBLE PRECISION";

    case java.sql.Types.LONGVARCHAR :

      return "TEXT";

    case java.sql.Types.BINARY :

      return "BINARY(" + field.getLength() + ")";

    case java.sql.Types.VARBINARY :

      return "VARBINARY(" + field.getLength() + ")";

    case java.sql.Types.LONGVARBINARY :

      return "IMAGE";

    case java.sql.Types.DATE :

    case java.sql.Types.TIME :

    case java.sql.Types.TIMESTAMP :

      return "DATETIME";

    case java.sql.Types.CHAR :

      return "CHAR(" + field.getLength() + ")";

    case java.sql.Types.VARCHAR :

      return "VARCHAR(" + field.getLength() + ")";

    case java.sql.Types.DECIMAL :

      return "DECIMAL(" + field.getLength() + ", " + field.getScale() + ")";

    case java.sql.Types.NUMERIC :

      return "NUMERIC(" + field.getLength() + ", " + field.getScale() + ")";

    case java.sql.Types.OTHER :

      return null;
    }

    return null;
  }

  /**
   * This method was created by a SmartGuide.
   * @return java.lang.String
   * @param field org.dbgen.Field
   */
  protected String getDatabaseSybaseType(Field field)
  {
    switch (field.getType())
    {
    case java.sql.Types.BIT :

      return "BIT";

    case java.sql.Types.TINYINT :

      return "TINYINT";

    case java.sql.Types.SMALLINT :

      return "SMALLINT";

    case java.sql.Types.INTEGER :

      return "INT";

    case java.sql.Types.REAL :

      return "REAL";

    case java.sql.Types.FLOAT :

      return "FLOAT";

    case java.sql.Types.DOUBLE :

      return "DOUBLE PRECISION";

    case java.sql.Types.LONGVARCHAR :

      return "TEXT";

    case java.sql.Types.BINARY :

      return "BINARY(" + field.getLength() + ")";

    case java.sql.Types.VARBINARY :

      return "VARBINARY(" + field.getLength() + ")";

    case java.sql.Types.LONGVARBINARY :

      return "IMAGE";

    case java.sql.Types.DATE :

    case java.sql.Types.TIME :

    case java.sql.Types.TIMESTAMP :

      return "DATETIME";

    case java.sql.Types.CHAR :

      return "CHAR(" + field.getLength() + ")";

    case java.sql.Types.VARCHAR :

      return "VARCHAR(" + field.getLength() + ")";

    case java.sql.Types.DECIMAL :

      return "DECIMAL(" + field.getLength() + ", " + field.getScale() + ")";

    case java.sql.Types.NUMERIC :

      return "NUMERIC(" + field.getLength() + ", " + field.getScale() + ")";

    case java.sql.Types.OTHER :

      return null;
    }

    return null;
  }

  /**
   * Get the value of a field (in Java syntax) used in a SQL statement.
   * @return java.lang.String
   * @param field org.dbgen.Field
   */
  protected String getFieldQuote(Field field)
  {
    String var = getFieldVariable(field);

    if (field.isNumberType() || field.isDecimalType())
    {
      return var;
    }
    else
      if (field.isBooleanType())
    {
      return "(" + var + " ? 1 : 0)";
    }
    else
      if (field.isDateType())
    {
      if (msqlTrick)
      {
        if (field.getType() == java.sql.Types.DATE)
        {
          var = "org.dbgen.support.MsqlHelper.formatDate(" + var + ")";
        }
        else
          if (field.getType() == java.sql.Types.TIME)
        {
          var = "org.dbgen.support.MsqlHelper.formatTime(" + var + ")";
        }
        else
        {
          var = "org.dbgen.support.MsqlHelper.formatTimestamp(" + var + ")";
        }
      }

      return "\"'\" + " + var + " + \"'\"";
    }
    else
    {
      return "\"'\" + org.dbgen.support.StringHelper.escape(" + var + ") + \"'\"";
    }
  }

  /**
   * Get the value of a field (in Java syntax) used in a SQL statement.
   * @return java.lang.String
   * @param field org.dbgen.Field
   */
  protected String getFieldQuote2(Field field, String var)
  {
    if (field.isNumberType() || field.isDecimalType())
    {
      return var;
    }
    else
      if (field.isBooleanType())
    {
      return "(" + var + " ? 1 : 0)";
    }
    else
      if (field.isDateType())
    {
      if (msqlTrick)
      {
        if (field.getType() == java.sql.Types.DATE)
        {
          var = "org.dbgen.support.MsqlHelper.formatDate(" + var + ")";
        }
        else
          if (field.getType() == java.sql.Types.TIME)
        {
          var = "org.dbgen.support.MsqlHelper.formatTime(" + var + ")";
        }
        else
        {
          var = "org.dbgen.support.MsqlHelper.formatTimestamp(" + var + ")";
        }
      }

      return "\"'\" + " + var + " + \"'\"";
    }
    else
    {
      return "\"'\" + org.dbgen.support.StringHelper.escape(" + var + ") + \"'\"";
    }
  }

  /**
   * Get the field variable used in the generated class.
   * @return java.lang.String
   * @param field org.dbgen.Field
   */
  protected String getFieldVariable(Field field)
  {
    return "_" + field.getName();
  }

  /**
   * Get the Java expression used to make a new Java object based on a primitive type.
   * @return java.lang.String
   * @param javaType int
   * @param val java.lang.String
   */
  protected String getJavaNewExpr(String javaType, String val)
  {
    if (javaType.indexOf(".") >= 0 || javaType.equals("String") || javaType.equals("byte[]"))
    {
      return val;
    }
    else
      if (javaType.equals("int"))
    {    // special case: int
      return "new Integer(" + val + ")";
    }
    else
    {
      return "new " + Output.capIt(javaType) + "(" + val + ")";
    }
  }

  /**
   * This method was created by a SmartGuide.
   * @return java.lang.String
   * @param javaType java.lang.String
   * @param val java.lang.String
   */
  protected String getJavaObjectValue(String javaType, String val)
  {
    // Raymond 3/23/98
    if (javaType.indexOf(".") >= 0 || javaType.equals("String") || javaType.equals("byte[]"))
    {
      return "(" + javaType + ") " + val;
    }
    else
      if (javaType.equals("int"))
    {    // special case: int
      return "((Integer) " + val + ").intValue()";
    }
    else
    {
      return "((" + Output.capIt(javaType) + ")" + val + ")." + javaType + "Value()";
    }
  }

  /**
   * Get the corresponding Java type from a SQL type.
   * @return java.lang.String
   * @param field org.dbgen.Field
   */
  protected String getJavaType(Field field)
  {
    int type = field.getType();

    switch (type)
    {
    case java.sql.Types.CHAR :

    case java.sql.Types.VARCHAR :

    case java.sql.Types.LONGVARCHAR :

      return "String";

    case java.sql.Types.BINARY :

    case java.sql.Types.VARBINARY :

    case java.sql.Types.LONGVARBINARY :

      return "byte[]";

    case java.sql.Types.NUMERIC :

    case java.sql.Types.DECIMAL :

      return "java.math.BigDecimal";

    case java.sql.Types.BIT :

      return "boolean";

    case java.sql.Types.TINYINT :

      return "byte";

    case java.sql.Types.SMALLINT :

      return "short";

    case java.sql.Types.INTEGER :

      return "int";

    case java.sql.Types.BIGINT :

      return "long";

    case java.sql.Types.REAL :

      return "float";

    case java.sql.Types.FLOAT :

    case java.sql.Types.DOUBLE :

      return "double";

    case java.sql.Types.DATE :

      return "java.sql.Date";

    case java.sql.Types.TIME :

      return "java.sql.Time";

    case java.sql.Types.TIMESTAMP :

      return "java.sql.Timestamp";

    case java.sql.Types.OTHER :

      return "Object";
    }

    return "** UNKNOWN **";
  }

  /**
   * Get the JDBC getXXX method name for the specified field.
   * @return java.lang.String
   * @param field org.dbgen.Field
   */
  protected String getJdbcGetMethod(Field field)
  {
    String r = getJavaType(field);

    /* special case for binary types */
    if (r.equals("byte[]"))
      return "getBytes";
    int idx = r.lastIndexOf(".");

    if (idx > 0)
    {
      r = r.substring(idx + 1, r.length());
    }

    return "get" + r.toUpperCase().charAt(0) + r.substring(1, r.length());
  }

  /**
   * This method was created by a SmartGuide.
   * @return java.lang.String
   * @param project org.dbgen.Project
   * @param table org.dbgen.Table
   */
  protected String getPackageName(Project project, Table table)
  {

    /* Retrieve default package name and table override package name. */
    String packageName = project.getDefaultPackageName();
    String overridePackageName = table.getPackageName();

    if (overridePackageName != null && !org.dbgen.support.StringHelper.isBlank(overridePackageName))
      packageName = overridePackageName;

    return packageName;
  }

  /**
   * This method was created by a SmartGuide.
   * @return java.lang.String
   * @param field org.dbgen.Field
   */
  public String getSQLType(Field field, int dbtype)
  {
    switch (dbtype)
    {
    case 0 :

      /* default SQL types */

      int type = field.getType();
      String s = FieldTypes.id2string(type);
      boolean hasLength = FieldTypes.getHasLength(type);
      boolean hasScale = FieldTypes.getHasScale(type);

      if (hasLength && hasScale)
        s += "(" + field.getLength() + "," + field.getScale() + ")";
      else
        if (hasLength)
        s += "(" + field.getLength() + ")";
      else
        if (hasScale)
        s += "(" + field.getScale() + ")";

      return s;

    case DATABASE_ORACLE :

      return getDatabaseOracleType(field);

    case DATABASE_SYBASE :

      return getDatabaseSybaseType(field);

    case DATABASE_ACCESS :

      return getDatabaseAccessType(field);

    case DATABASE_MSQL :

      return getDatabaseMsqlType(field);

    case DATABASE_SQLSERVER :

      return getDatabaseSqlServerType(field);
    default :

      return "UNKNOWN TYPE";
    }
  }

  /**
   * Get the parsing method (string -> Java) for the specified field.
   * @return java.lang.String
   * @param field org.dbgen.Field
   */
  protected String getStringParseMethod(Field field)
  {
    int type = field.getType();

    switch (type)
    {
      /*
       case java.sql.Types.CHAR:
       case java.sql.Types.VARCHAR:
       case java.sql.Types.LONGVARCHAR:
       return "String";
       */
    case java.sql.Types.NUMERIC :

    case java.sql.Types.DECIMAL :

      return "new java.math.BigDecimal";

    case java.sql.Types.BIT :

      return "\"true\".equals";

    case java.sql.Types.TINYINT :

      return "Byte.parseByte";

    case java.sql.Types.SMALLINT :

      return "Short.parseShort";

    case java.sql.Types.INTEGER :

      return "Integer.parseInt";

    case java.sql.Types.BIGINT :

      return "Long.parseLong";

    case java.sql.Types.REAL :

      return "org.dbgen.support.Parsing.parseFloat";

    case java.sql.Types.FLOAT :

    case java.sql.Types.DOUBLE :

      return "org.dbgen.support.Parsing.parseDouble";

    case java.sql.Types.DATE :

      return "java.sql.Date.valueOf";

    case java.sql.Types.TIME :

      return "java.sql.Time.valueOf";

    case java.sql.Types.TIMESTAMP :

      return "java.sql.Timestamp.valueOf";
    }

    return "** UNKNOWN **";
  }

  /**
   * This method was created by a SmartGuide.
   * @return java.util.Vector
   */
  public static Vector getSupportedDatabases()
  {
    Vector db = new Vector();
    db.addElement(new Map(DATABASE_ORACLE, "Oracle"));
    db.addElement(new Map(DATABASE_SYBASE, "Sybase"));
    db.addElement(new Map(DATABASE_MSQL, "mSQL"));
    db.addElement(new Map(DATABASE_ACCESS, "Microsoft Access"));
    db.addElement(new Map(DATABASE_SQLSERVER, "Microsoft SQL Sever"));

    return db;
  }

  /**
   * Make the clear() method.
   * @param table org.dbgen.Table
   */
  protected void makeClearMethod(Table table)
  {
    output.println();
    output.comment(true, "Clear the " + table.getClassName() + " object.");
    output.iprintln("public void clear()");
    output.iprintln("{");
    output.indent();
    Vector fields = table.getFields();

    for (int i = 0; i < fields.size(); i++)
    {
      Field f = (Field) fields.elementAt(i);
      String var = getFieldVariable(f);
      output.iprint(var + " = ");

      if (f.isNumberType())
      {
        output.println("0;");
      }
      else
        if (f.isBooleanType())
      {
        output.println("false;");
      }
      else
      {
        output.println("null;");
      }
    }

    output.deindent();
    output.iprintln("}");

    return;
  }

  /**
   * Make a constructor.
   * @param comment java.lang.String
   * @param tableName java.lang.String
   * @param fields java.util.Vector
   */
  protected void makeConstructorHelper(String comment, String tableName, Vector fields)
  {
    output.iprintln("");
    output.comment(true, comment);
    output.iprint("public " + tableName + "(");

    for (int i = 0, j = 0; i < fields.size(); i++)
    {
      Field f = (Field) fields.elementAt(i);

      if (j > 0)
      {
        output.print(", ");
      }

      output.print(getJavaType(f) + " " + getFieldVariable(f));
      j++;
    }

    output.println(")");
    output.iprintln("{");
    output.indent();

    for (int i = 0; i < fields.size(); i++)
    {
      String var = getFieldVariable((Field) fields.elementAt(i));
      output.iprintln("this." + var + " = " + var + ";");
    }

    output.deindent();
    output.iprintln("}");

    return;
  }

  /**
   * Make all necessary constructors.
   * @param table org.dbgen.Table
   */
  protected void makeConstructors(Table table)
  {
    String className = table.getClassName();
    output.iprintln("");
    output.comment(true, "Default Constructor");
    output.iprintln("public " + className + "()");
    output.iprintln("{");
    output.iprintln("}");
    Vector keyFields = table.getKeyFields();
    Vector fields = table.getFields();
    makeConstructorHelper("Key Constructor", className, keyFields);

    if (keyFields.size() != fields.size())
      makeConstructorHelper("Normal Constructor", className, fields);

    return;
  }

  /**
   * This method can be overriden in the subclass so that the generator
   * can be customized for special use.
   * @param table org.dbgen.Table
   */
  public void makeCustomMethods(Table table)
  {
    return;
  }

  /**
   * Make the field definitions.
   * @param table org.dbgen.Table
   */
  protected void makeFieldDefinitions(Table table)
  {
    output.comment(false, "Table Fields");
    Vector fields = table.getFields();

    for (int i = 0; i < fields.size(); i++)
    {
      Field f = (Field) fields.elementAt(i);
      output.iprintln("private " + getJavaType(f) + " " + getFieldVariable(f) + ";");
    }

    /*
     // For new/changed/updated flag
     output.iprintln("private int __flag = FLAG_NEW;");
     */

    /* For property change event */
    output.iprintln("protected transient java.beans.PropertyChangeSupport propertyChange = new java.beans.PropertyChangeSupport(this);");

    return;
  }

  public void makeGetFieldMethod(Table table)
  {
    // Raymond 3/23/98

    output.println();
    output.comment(true, "Get data " + table.getClassName() + " by object array.");
    output.iprintln("public Object[] getFields()");
    output.iprintln("{");
    output.indent();
    Vector fields = table.getFields();
    output.iprintln("Object[] objects = new Object[" + fields.size() + "];");

    for (int i = 0; i < fields.size(); i++)
    {
      Field f = (Field) fields.elementAt(i);
      output.iprintln("objects[" + i + "] = " + getJavaNewExpr(getJavaType(f), "_" + f.getName()) + ";");
    }

    output.iprintln("return objects;");
    output.deindent();
    output.iprintln("}");
  }

  /**
   * Make the getInsertValues() method.
   * @param table org.dbgen.Table
   */
  protected void makeGetInsertValuesMethod(Table table)
  {
    output.println();
    output.comment(true, "Get the INSERT values clause.");
    output.iprintln("public String getInsertValues()");
    output.iprintln("{");
    output.indent();
    output.iprint("return ");
    Vector fields = table.getFields();

    for (int i = 0; i < fields.size(); i++)
    {
      Field f = (Field) fields.elementAt(i);
      String varq = getFieldQuote(f);

      if (i > 0)
      {
        output.print(" + \",\" + ");
      }

      output.print(varq);
    }

    output.println(";");
    output.deindent();
    output.iprintln("}");

    return;
  }

  /**
   * Make the getKeyConditions() method.
   * @param table org.dbgen.Table
   */
  protected void makeGetKeyConditionsForRelationMethod(Table table)
  {

    // Generates one method for each master table

    Project thisProject = table.getProject();
    Vector relations = thisProject.getRelationsByForeignTable(table.getTableName());
    Relation r = null;
    String curMasterTableName = "";
    output.iprintln("public String getKeyConditions( Data masterData )");
    output.println();
    output.iprintln("{");
    output.indent();
    int i = 0;

    while (i < relations.size())
    {
      r = (Relation) relations.elementAt(i);
      curMasterTableName = r.getMasterTable().getTableName();

      // header
      output.println();
      output.comment(true, "Test case for master table to this foreign table");
      output.iprintln("if (masterData instanceof " + r.getMasterTable().getClassName() + ")");
      output.iprint("\treturn ");
      int fieldCount = 0;

      while (curMasterTableName.equals(r.getMasterTable().getTableName()))
      {
        Field fk = r.getForeignKey();
        Field mk = r.getMasterKey();

        if (fieldCount > 0)
          output.print(" + \" AND \" + ");
        output.print("\"" + fk.getName() + " = \" + " +

          // add quote if necessary
          getFieldQuote2(mk, "((" + r.getMasterTable().getClassName() + ") masterData).get" + mk.getName() + "()"));
        i++;
        fieldCount++;

        if (i >= relations.size())
        break;
        r = (Relation) relations.elementAt(i);
      }

      output.println(";");
      output.println();
    }

    // Should (at least) generate this empty method for interface

    output.iprintln("return \"\";");
    output.deindent();
    output.iprintln("}");

    return;
  }

  /**
   * Make the getKeyConditions() method.
   * @param table org.dbgen.Table
   */
  protected void makeGetKeyConditionsMethod(Table table)
  {
    output.println();
    output.comment(true, "Get the SQL condition for key fields.");
    output.iprintln("public String getKeyConditions()");
    output.iprintln("{");
    output.indent();
    output.iprint("return ");
    Vector keyFields = table.getKeyFields();

    for (int i = 0; i < keyFields.size(); i++)
    {
      Field f = (Field) keyFields.elementAt(i);
      String var = getFieldVariable(f);
      String varq = getFieldQuote(f);

      if (i > 0)
      {
        output.print(" + \" AND \" + ");
      }

      output.print("\"" + f.getName() + " = \" + " + varq);
    }

    output.println(";");
    output.deindent();
    output.iprintln("}");

    return;
  }

  /**
   * Make all getXXX/setXXX methods.
   * @param table org.dbgen.Table
   */
  protected void makeGetSetMethods(Table table)
  {
    Vector fields = table.getFields();

    for (int i = 0; i < fields.size(); i++)
    {
      Field f = (Field) fields.elementAt(i);
      String javaType = getJavaType(f);
      String var = getFieldVariable(f);
      String name = f.getName();
      String Name = Output.capIt(name);
      output.println();
      output.comment(true, "Get the " + name + " property (" + javaType + ") value.");
      output.iprintln("public " + javaType + " get" + Name + "()");
      output.iprintln("{");
      output.indent();
      output.iprintln("return " + var + ";");
      output.deindent();
      output.iprintln("}");
      output.println();
      output.comment(true, "Set the " + name + " property (" + javaType + ") value.");
      output.iprintln("public void set" + Name + "(" + javaType + " " + var + ")");
      output.iprintln("{");
      output.indent();
      output.iprintln("/* Get the old property value for fire property change event. */");

      // Fixed firing event  Raymond 11/19/98
      output.iprintln(javaType + " oldValue = this." + var + ";");
      output.iprintln("/* Set the " + name + " property (attribute) to the new value. */");
      output.iprintln("this." + var + " = " + var + ";");
      output.iprintln("/* Fire (signal/notify) the " + name + " property change event. */");
      output.iprintln("firePropertyChange(\"" + name + "\", " + getJavaNewExpr(javaType, "oldValue") + ", " + getJavaNewExpr(javaType, var) + ");");
      output.deindent();
      output.iprintln("}");
    }

    /*
     // Generate get/set method for flag
     output.println();
     output.comment(true, "Get the db data's flag.");
     output.iprintln("public int get__Flag()");
     output.iprintln("{");
     output.indent();
     output.iprintln("return __flag;");
     output.deindent();
     output.iprintln("}");
     
     output.println();
     output.comment(true, "Set the db data's flag.");
     output.iprintln("public void set__Flag( int flag )");
     output.iprintln("{");
     output.indent();
     output.iprintln("__flag = flag;");
     output.deindent();
     output.iprintln("}");
     */

    return;
  }

  /**
   * Meke the getTableName() method.
   * @return java.lang.String
   * @param table org.dbgen.Table
   */
  protected String makeGetTableNameMethod(Table table)
  {
    output.println();
    output.comment(true, "Get the table name.");
    output.iprintln("public String getTableName()");
    output.iprintln("{");
    output.indent();
    output.iprintln("return \"" + table.getTableName() + "\";");
    output.deindent();
    output.iprintln("}");

    return null;
  }

  /**
   * Make the getUpdateValues() method.
   * @param table org.dbgen.Table
   */
  protected void makeGetUpdateValuesMethod(Table table)
  {
    output.println();
    output.comment(true, "Get the UPDATE set clause.");
    output.iprintln("public String getUpdateValues()");
    output.iprintln("{");
    output.indent();
    output.iprint("return ");
    Vector fields = table.getFields();
    int count = 0;

    for (int i = 0; i < fields.size(); i++)
    {
      Field f = (Field) fields.elementAt(i);
      String varq = getFieldQuote(f);

      if (!f.getIsReadOnly())
      {
        if (count > 0)
        {
          output.print(" + \",\" + ");
        }

        output.print("\"" + f.getName() + "\" + \" = \" + " + varq);
        count++;
      }
    }

    if (count <= 0)
      output.print("\"\"");
    output.println(";");
    output.deindent();
    output.iprintln("}");

    return;
  }

  /**
   * Make the header for Data class.
   * @param table org.dbgen.Table
   */
  protected void makeHeader(Table table)
  {
    output.comment(true, "Class " + table.getClassName() + ".java - for table <" + table.getTableName() + ">" + NL + NL + "THIS PROGRAM IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND. " + "YOU MAY FREELY USE AND DISTRIBUTE THE GENERATED JAVA SOURCE " + "FROM THIS PROGRAM ROYALTY FREE. " + NL + NL + "Generated by: " + NAME + NL + "Generated on: " + Util.getDateString() + NL + NL + "Please visit our web site at " + "http://dbgen.sourceforge.net/.");

    return;
  }

  /**
   * Make the package/import statements.
   * @param table org.dbgen.Table
   */
  protected void makeImport(Project project, Table table)
  {

    /* Retrieve default package name and table override package name. */
    String packageName = getPackageName(project, table);

    if (packageName != null && !org.dbgen.support.StringHelper.isBlank(packageName))
    {
      output.println();
      output.println("package " + packageName + ";");
    }

    output.println();
    output.println("import java.sql.*;");
    output.println("import java.util.*;");
    output.println("import java.net.*;");
    output.println("import java.io.*;");
    output.println("import org.dbgen.support.*;");

    return;
  }

  /**
   * Make the makeObject() method.
   * @param table org.dbgen.Table
   */
  protected void makeMakeObjectMethod(Table table)
  {
    output.println();
    output.comment(true, "Make a new " + table.getClassName() + " object." + NL + "@exception SQLException When a SQL Exception occurred.");
    output.iprintln("public org.dbgen.support.Data makeObject(ResultSet rs)");
    output.iprintln(" throws SQLException");
    output.iprintln("{");
    output.indent();
    output.iprint("return new " + table.getClassName() + "(");
    Vector fields = table.getFields();

    for (int i = 0; i < fields.size(); i++)
    {
      Field f = (Field) fields.elementAt(i);

      if (i > 0)
      {
        output.print(", ");
      }

      if (f.isDecimalType())
      {
        output.print("rs." + getJdbcGetMethod(f) + "(" + (i + 1) + ", " + f.getScale() + ")");
      }
      else
      {
        output.print("rs." + getJdbcGetMethod(f) + "(" + (i + 1) + ")");
      }
    }

    output.println(");");
    output.deindent();
    output.iprintln("}");

    return;
  }

  /**
   * Make the property change support methods.
   */
  protected void makePropertyChangeMethods()
  {
    output.println();
    output.comment(true, "Add a property change listener");
    output.iprintln("public synchronized void addPropertyChangeListener(java.beans.PropertyChangeListener listener)");
    output.iprintln("{");
    output.indent();
    output.iprintln("if (propertyChange == null)");
    output.iprintln("\tpropertyChange = new java.beans.PropertyChangeSupport(this);");
    output.iprintln("propertyChange.addPropertyChangeListener(listener);");
    output.deindent();
    output.iprintln("}");
    output.println();
    output.comment(true, "Remove a property change listener");
    output.iprintln("public synchronized void removePropertyChangeListener(java.beans.PropertyChangeListener listener)");
    output.iprintln("{");
    output.indent();
    output.iprintln("if (propertyChange == null)");
    output.iprintln("\tpropertyChange = new java.beans.PropertyChangeSupport(this);");
    output.iprintln("propertyChange.removePropertyChangeListener(listener);");
    output.deindent();
    output.iprintln("}");
    output.println();
    output.comment(true, "Fire a property change event.");
    output.iprintln("public void firePropertyChange(String propertyName, Object oldValue, Object newValue)");
    output.iprintln("{");
    output.indent();
    output.iprintln("if (propertyChange == null)");
    output.iprintln("\tpropertyChange = new java.beans.PropertyChangeSupport(this);");
    output.iprintln("propertyChange.firePropertyChange(propertyName, oldValue, newValue);");
    output.deindent();
    output.iprintln("}");

    return;
  }

  /**
   * This method was created by a SmartGuide.
   * @param table org.dbgen.Table
   */
  public void makeSetFieldMethod(Table table)
  {
    // Raymond 3/23/98

    output.println();
    output.comment(true, "Set fields by object array.");
    output.iprintln("public void setFields(Object[] objects)");
    output.iprintln("{");
    output.indent();
    Vector fields = table.getFields();

    for (int i = 0; i < fields.size(); i++)
    {
      Field f = (Field) fields.elementAt(i);
      output.iprintln("_" + f.getName() + " = " + getJavaObjectValue(getJavaType(f), "objects[" + i + "]") + ";");
    }

    output.deindent();
    output.iprintln("}");
  }

  /**
   * This method was created by a SmartGuide.
   */
  protected void makeSetKeysFromParentMethod(Project project, Table table)
  {
    output.println();
    output.comment(true, "Set the key fields from a parent class.");
    output.iprintln("public void setKeysFromParent(org.dbgen.support.Data parent)");
    output.iprintln("{");
    output.indent();

    /*
     Vector parentRelations = project.getParentRelations(table);
     for (int i = 0; i < parentRelations.size(); i++)
     {
     Relation r = (Relation) parentRelations.elementAt(i);
     Vector keyMap = r.getKeyMap();
     String parentClass = r.getParent().getClassName();
     output.iprintln("if (parent instanceof " + parentClass + ") {");
     for (int j = 0; j < keyMap.size(); j++)
     {
     String[] fields = (String[]) keyMap.elementAt(j);
     output.iprintln("  set" + Output.capIt(fields[1]) + "(((" + parentClass + ")parent).get" + Output.capIt(fields[0]) + "());");
     }
     output.iprintln("}");
     }
     */
    output.deindent();
    output.iprintln("}");

    return;
  }

  /**
   * Make all necessary constructors.
   * @param table org.dbgen.Table
   */
  protected void makeTMConstructors(Table table)
  {
    String className = table.getClassName();
    output.iprintln("");
    output.comment(true, "Default Constructor");
    output.iprintln("public " + className + "DbGenTableModel()");
    output.iprintln("{");
    output.iprintln("\tsuper();");
    output.iprintln("}");

    return;
  }

  /**
   * Make the field definitions.
   * @param table org.dbgen.Table
   */
  protected void makeTMFieldDefinitions(Table table)
  {
    output.iprintln("static {");
    output.indent();
    output.iprint("setColumnNames(new String[] {");
    output.indent();
    Vector fields = table.getFields();

    for (int i = 0; i < fields.size(); i++)
    {
      Field f = (Field) fields.elementAt(i);
      output.print("\t\t\"" + f.getName() + "\"");

      if (i < fields.size() - 1)
      {
        output.println(",");
      }
    }

    output.deindent();
    output.iprintln(" });");
    output.deindent();
    output.iprintln("}");

    return;
  }

  /**
   * Make all getXXX/setXXX methods.
   * @param table org.dbgen.Table
   */

  protected void makeTMGetSetMethods(Table table)
  {


    // Gen createDbData method
    output.comment(true, "@return org.dbgen.support.Data");
    output.iprintln("public org.dbgen.support.Data createDbData() {");
    output.iprintln("\treturn new " + table.getClassName() + "();");
    output.iprintln("}");
    output.comment(true, "* Overrides AbstractTableModel method" + " * @return java.lang.Object" + " * @param row int" + " * @param column int" + " expecting type name");


    String className = table.getClassName();

    output.iprintln("public Object getValueAt(int row, int column) {");
    output.indent();
    output.iprintln(className + " data = (" + className + ") getDbDataRowAt( row);");
    output.iprintln("switch( column) {");
    output.indent();
    Vector fields = table.getFields();

    for (int i = 0; i < fields.size(); i++)
    {
      Field f = (Field) fields.elementAt(i);
      String javaType = getJavaType(f);
      String var = getFieldVariable(f);
      String name = f.getName();
      String Name = Output.capIt(name);
      output.iprintln("case " + i + ":");
      output.iprintln("\treturn " + getJavaNewExpr(javaType, "data.get" + name + "()") + ";");
    }

    output.iprintln("default:");
    output.iprintln("\treturn \"Unknown\";");
    output.deindent();
    output.iprintln("}");
    output.deindent();
    output.iprintln("}");


    output.comment(true, "Overrides AbstractTableModel method" +
      "@param aValue java.lang.Object"+
      "@param row int"+
      "@param column int"
      );

    output.iprintln("public void setValueAt(Object aValue, int row, int column) {");
    output.indent();
    output.iprintln(className + " data = (" + className + ") getDbDataRowAt( row);");
    output.iprintln("switch( column) {");
    output.indent();
    fields = table.getFields();

    for (int i = 0; i < fields.size(); i++)
    {
      Field f = (Field) fields.elementAt(i);
      String javaType = getJavaType(f);
      String var = getFieldVariable(f);
      String name = f.getName();
      String Name = Output.capIt(name);
      output.iprintln("case " + i + ":");
      output.iprintln("\tdata.set" + name + "(" + getJavaObjectValue(javaType, "aValue") + ");" );
      output.iprintln("\tbreak;");
    }

    output.iprintln("default:");
    output.deindent();
    output.iprintln("}");
    output.iprintln("");
    output.iprintln("fireTableChanged(new TableModelEvent(this, row, row, column));" );
    output.deindent();
    output.iprintln("}");
  }

  /**
   * Make the header for Data class.
   * @param table org.dbgen.Table
   */
  protected void makeTMHeader(Table table)
  {
    output.comment(true, "Class " + table.getClassName() + "DbGenTableModel.java - for table <" + table.getTableName() + ">" + NL + NL + "THIS PROGRAM IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND. " + "YOU MAY FREELY USE AND DISTRIBUTE THE GENERATED JAVA SOURCE " + "FROM THIS PROGRAM ROYALTY FREE. " + NL + NL + "Generated by: " + NAME + NL + "Generated on: " + Util.getDateString() + NL + NL + "Please visit our web site at " + "http://dbgen.sourceforge.net/.");
    output.println();
    output.comment(true,
      "This class inherits AbstractDbGenTableModel which subclasses" + NL +
      "Swing's AbstractTableModel." + NL + NL +
      "See main() for sample use." + NL + NL +
      "Functions in class AbstractDbGenTableModel for working with database:" + NL +
      "_addDbData( org.dbgen.support.Data data )" + NL +
      "_insertDbDataAt( int row,org.dbgen.support.Data data)" + NL +
      "_changeDbDataAt( int row )" + NL +
      "_deleteDbDataAt( int row )" + NL +
      "_commitDbData( )" + NL
      );

    return;
  }

  /**
   * Make the package/import statements.
   * @param table org.dbgen.Table
   */
  protected void makeTMImport(Project project, Table table, boolean isJavax )
  {

    /* Retrieve default package name and table override package name. */
    String packageName = getPackageName(project, table);

    if (packageName != null && !org.dbgen.support.StringHelper.isBlank(packageName))
    {
      output.println();
      output.println("package " + packageName + ";");
    }

    output.println();

    if ( isJavax )
    {
      output.println("import javax.swing.*;");
      output.println("import javax.swing.event.*;");
      output.println("import javax.swing.table.*;");
    }
    else
    {
      output.println("import javax.swing.*;");
      output.println("import javax.swing.event.*;");
      output.println("import javax.swing.table.*;");
    }

    output.println("import org.dbgen.support.*;");

    return;
  }

  /**
   * Make all necessary constructors.
   * @param table org.dbgen.Table
   */
  protected void makeTMMain(Table table)
  {

    String className = table.getClassName() + "DbGenTableModel";
    String dbDataClassName = table.getClassName();

    output.iprintln("public static void main(String args[]) {" );
    output.indent();
    output.iprintln("try {" );
    output.indent();

    // create components
    output.iprintln("JFrame f = new JFrame();" );
    output.iprintln("final " + className + " tm =" );
    output.iprintln(" new " + className + "() {" );
    output.iprintln(" public boolean isCellEditable( int row, int column ) {" );
    output.iprintln(" return true;" );
    output.iprintln(" }" );
    output.iprintln(" };" );

    output.iprintln("final JTable tbl =" );
    output.iprintln(" new JTable(tm) {" );
    output.iprintln(" public void editingStopped( ChangeEvent ev ) {" );
    output.iprintln(" tm._changeDbDataAt( getSelectedRow());" );
    output.iprintln(" super.editingStopped(ev);" );
    output.iprintln(" }" );
    output.iprintln(" };" );

    output.iprintln(" tbl.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);" );

    output.iprintln(" TableColumnModel tcm = tbl.getColumnModel();" );
    output.iprintln(" for ( int i=0; i<tbl.getColumnCount(); i++ ) {" );
    output.iprintln(" tcm.getColumn(i).setMinWidth(100);" );
    output.iprintln(" }" );

    output.iprintln("JPanel pnlButtons = new JPanel( new java.awt.GridLayout() );" );

    output.comment(true, "create buttons");
    output.iprintln("JButton btnAdd = new JButton(\"Add\");" );
    output.iprintln("JButton btnDelete = new JButton(\"Delete\");" );
    output.iprintln("JButton btnCommit = new JButton(\"Commit\");" );
    output.iprintln("JButton btnClose = new JButton(\"Close\");" );

    output.iprintln("btnAdd.addActionListener(" );
    output.iprintln(" new java.awt.event.ActionListener() {" );
    output.iprintln(" public void actionPerformed( java.awt.event.ActionEvent ev ) {" );
    output.iprintln(" tm._addDbData( new " + dbDataClassName + "());" );
    output.iprintln(" }" );
    output.iprintln(" }" );
    output.iprintln(");" );

    output.iprintln("btnDelete.addActionListener(" );
    output.iprintln(" new java.awt.event.ActionListener() {" );
    output.iprintln(" public void actionPerformed( java.awt.event.ActionEvent ev ) {" );
    output.iprintln(" tm._deleteDbDataAt( tbl.getSelectedRow());" );
    output.iprintln(" }" );
    output.iprintln(" }" );
    output.iprintln(");" );

    output.iprintln("btnCommit.addActionListener(" );
    output.iprintln(" new java.awt.event.ActionListener() {" );
    output.iprintln(" public void actionPerformed( java.awt.event.ActionEvent ev ) {" );
    output.iprintln(" tm._commitDbData();" );
    output.iprintln(" }" );
    output.iprintln(" }" );
    output.iprintln(");" );

    output.iprintln("btnClose.addActionListener(" );
    output.iprintln(" new java.awt.event.ActionListener() {" );
    output.iprintln(" public void actionPerformed( java.awt.event.ActionEvent ev ) {" );
    output.iprintln(" System.exit(0);" );
    output.iprintln(" }" );
    output.iprintln(" }" );
    output.iprintln(");" );

    output.iprintln("pnlButtons.add(btnAdd);" );
    output.iprintln("pnlButtons.add(btnDelete);" );
    output.iprintln("pnlButtons.add(btnCommit);" );
    output.iprintln("pnlButtons.add(btnClose);" );

    output.comment(true, "setup for database access" );
    output.iprintln("String drv = \"" + DbGenPro.getValue(DbGenPro.JDBC_DRIVER) + "\";" );
    output.iprintln("String url = \"" + DbGenPro.getValue(DbGenPro.JDBC_URL) + "\";" );
    output.iprintln("String usr = \"" + DbGenPro.getValue(DbGenPro.JDBC_USERNAME) + "\";" );
    output.iprintln("String pwd = \"" + DbGenPro.getValue(DbGenPro.JDBC_PASSWORD) + "\";" );
    output.iprintln("DbAccess dba = new org.dbgen.support.DbAccess(drv, url, usr, pwd);");
    output.iprintln("Driver driver = new Driver(dba);" );
    output.iprintln("dba.setDebug(true); // Set to TRUE for verbose SQL inspection" );
    output.iprintln("tm.setDbGenDriver(driver);" );
    output.iprintln("tm._retrieveDbData(\"\",\"\");" );

    output.comment(true, "show frame");
    output.iprintln("f.getContentPane().add( \"Center\", new JScrollPane(tbl));" );
    output.iprintln("f.getContentPane().add( \"South\", pnlButtons);" );
    output.iprintln("f.pack();" );
    output.iprintln("f.show();" );

    output.deindent();
    output.iprintln("} catch ( Exception ex ) {" );
    output.iprintln(" ex.printStackTrace();" );
    output.iprintln("}" );
    output.deindent();
    output.iprintln("}" );
  }

  /**
   * This method was created by a SmartGuide.
   * @param table org.dbgen.Table
   */
  protected void makeToStringMethod(Table table)
  {
    output.println();
    output.comment(true, "Return a printable representation of the " + table.getClassName() + " object.");
    output.iprintln("public String toString()");
    output.iprintln("{");
    output.indent();
    output.iprint("return \"#(\" + ");
    Vector fields = table.getFields();

    for (int i = 0; i < fields.size(); i++)
    {
      Field f = (Field) fields.elementAt(i);
      String var = getFieldVariable(f);

      if (i > 0)
      {
        output.print(" + \" \" + ");
      }

      output.print(var);
    }

    output.println(" + \")\";");
    output.deindent();
    output.iprintln("}");

    return;
  }

  /**
   * Sets the writer property (java.io.PrintWriter) value.
   * @param writer The new value for the property.
   */
  public void setWriter(java.io.PrintWriter writer)
  {
    fieldWriter = writer;
    output.setWriter(writer);

    return;
  }

  /**
   * Testing.
   * @param args java.lang.String[]
   */
  public static void test(String args[])
  {
    DbGen generator = new DbGen();
    Table table = new Table();
    table.setClassName("Zip");
    table.setTableName("Zip");
    table.addField(new Field(true, "zip_code", java.sql.Types.VARCHAR, 10, 0));
    table.addField(new Field(false, "state_code", java.sql.Types.CHAR, 2, 0));
    table.addField(new Field(false, "city_name", java.sql.Types.VARCHAR, 30, 0));
    table.addField(new Field(false, "county_name", java.sql.Types.VARCHAR, 30, 0));
    table.addField(new Field(false, "county_code", java.sql.Types.VARCHAR, 4, 0));
    Table table2 = new Table();
    table2.setClassName("Customer");
    table2.setTableName("Customer");
    table2.addField(new Field(true, "SSN", java.sql.Types.CHAR, 9, 0));
    table2.addField(new Field(false, "FirstName", java.sql.Types.VARCHAR, 20, 0));
    table2.addField(new Field(false, "LastName", java.sql.Types.VARCHAR, 20, 0));
    table2.addField(new Field(false, "Age", java.sql.Types.TINYINT, 0, 0));
    table2.addField(new Field(false, "DateOfBirth", java.sql.Types.DATE, 0, 0));
    table2.addField(new Field(false, "RegisterTimestamp", java.sql.Types.TIMESTAMP, 0, 0));
    table2.addField(new Field(false, "Male", java.sql.Types.BIT, 0, 0));
    table2.addField(new Field(false, "Zip", java.sql.Types.VARCHAR, 10, 0));
    Project project = new Project();
    project.setOutputDirectory("d:\\tmp");
    project.setDefaultPackageName("org.dbgen.test");
    Hashtable options = new Hashtable();
    options.put(GENERATE_CREATE_SCRIPT, new Boolean(true));
    options.put(GENERATE_DATA_CLASS, new Boolean(true));
    options.put(GENERATE_TEST_CLASS, new Boolean(true));

    //	options.put(DATABASE_TYPE, new Integer(DATABASE_SYBASE));
    //	options.put(DATABASE_TYPE, new Integer(DATABASE_ACCESS));
    options.put(DATABASE_TYPE, new Integer(DATABASE_MSQL));

    //	options.put(DATABASE_TYPE, new Integer(DATABASE_ORACLE));

    try
    {
      generator.generate(project, table, options);
      generator.generate(project, table2, options);
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }

    return;
  }
}
